"""
Methods for computing eigenvalues and eigenvectors of sparse matrices.
"""
import numpy as np
import scipy as sp
import scipy.sparse
from typing import Tuple

from . import utility
from . import stag_internal


def compute_eigensystem(mat: scipy.sparse.spmatrix,
                        num: int,
                        which: str = 'SM') -> Tuple[np.ndarray, np.ndarray]:
    r"""
    Compute the eigenvalues and eigenvectors of a given matrix.

    By default, this will compute the eigenvalues of smallest magnitude.
    This default can be overridden by the `which `parameter which takes a
    string and should take one of the following values.
      - `SM` will return the eigenvalues with smallest magnitude
      - `LM` will return the eigenvalues with largest magnitude

    The following example demonstrates how to compute the 3 largest eigenvectors
    and eigenvalues of a cycle graph.

    \code{.py}
        import stag.graph
        import stag.spectrum

        myGraph = stag.graph.cycle_graph(10)
        lap = myGraph.normalised_laplacian()
        eigenvalues, eigenvectors = stag.spectrum.compute_eigensystem(
            lap, 3, 'LM')
    \endcode

    @param mat the matrix on which to operate
    @param num the number of eigenvalues and eigenvectors to compute
    @param which (optional) a string indicating which eigenvectors to calculate
    @returns a tuple containing the computed eigenvalues and eigenvectors
    """
    return sp.sparse.linalg.eigs(mat, k=num, which=which)


def compute_eigenvalues(mat: scipy.sparse.spmatrix,
                        num: int,
                        which: str = 'SM') -> np.ndarray:
    r"""
    Compute the eigenvalues of a given matrix.

    By default, this will compute the eigenvalues of smallest magnitude.
    This default can be overridden by the which parameter which takes a string
    and should be one of the following.
      - `SM` will return the eigenvalues with smallest magnitude
      - `LM` will return the eigenvalues with largest magnitude

    If you would like to calculate the eigenvectors and eigenvalues together, then
    you should instead use stag.spectrum.compute_eigensystem.

    @param mat the matrix on which to operate
    @param num the number of eigenvalues to compute
    @param which (optional) a string indicating which eigenvalues to calculate
    @returns a numpy array containing the computed eigenvalues
    """
    eigs, _ = compute_eigensystem(mat, num, which=which)
    return eigs

def compute_eigenvectors(mat: scipy.sparse.spmatrix,
                         num: int,
                         which: str = 'SM') -> np.ndarray:
    r"""
    Compute the eigenvectors of a given matrix.

    By default, this will compute the eigenvectors corresponding to the
    eigenvalues of smallest magnitude.
    This default can be overridden by the which parameter which takes a string
    and should be one of the following.
      - `SM` will return the vectors corresponding to eigenvalues with smallest magnitude
      - `LM` will return the vectors corresponding to eigenvalues with largest magnitude

    If you would like to calculate the eigenvectors and eigenvalues together, then
    you should instead use stag.spectrum.compute_eigensystem.

    @param mat the matrix on which to operate
    @param num the number of eigenvectors to compute
    @param which (optional) a string indicating which eigenvectors to calculate
    @returns a numpy array containing the computed eigenvectors as columns
    """
    _, eigvecs = compute_eigensystem(mat, num, which=which)
    return eigvecs


def power_method(mat: scipy.sparse.spmatrix,
                 num_iterations: int = None,
                 initial_vector: np.ndarray = None) -> np.ndarray:
    r"""
    Apply the power method to compute the dominant eigenvector of a matrix.

    Given a matrix \f$M\f$, an initial vector \f$v_0\f$, and a number of
    iterations \f$t\f$, the power method calculates the vector

    \f[
       v_t = M^t v_0,
    \f]

    which is close to the eigenvector of \f$M\f$ with largest eigenvalue.

    The running time of the power method is \f$O(t \cdot \mathrm{nnz}(M))\f$, where
    \f$\mathrm{nnz}(M)\f$ is the number of non-zero elements in the matrix \f$M\f$.

    @param mat the matrix \f$M\f$ on which to operate.
    @param num_iterations (optional) the number of iterations of the power
                          method to apply. It this argument is omitted,
                          \f$O(\log(n))\f$ iterations are used which results
                          in a vector whose Rayleigh quotient is a \f$(1 - \epsilon)\f$
                          approximation of the dominant eigenvalue.
    @param initial_vector (optional) the initial vector to use for the power
                          iteration. If this argument is omitted, a random unit
                          vector will be used.
    @return the vector \f$v_t\f$ computed by repeated multiplication with \f$M\f$.
    """
    if num_iterations is None:
        if initial_vector is None:
            return stag_internal.power_method(utility.scipy_to_swig_sprs(mat))
        else:
            return stag_internal.power_method(utility.scipy_to_swig_sprs(mat),
                                              initial_vector.astype(float))
    elif initial_vector is None:
        return stag_internal.power_method(utility.scipy_to_swig_sprs(mat),
                                          num_iterations)
    else:
        return stag_internal.power_method(utility.scipy_to_swig_sprs(mat),
                                          num_iterations, initial_vector.astype(float))


def rayleigh_quotient(mat: scipy.sparse.spmatrix, vec: np.ndarray) -> float:
    r"""
    Compute the Rayleigh quotient of the given vector and matrix.

    Given a matrix \f$M\f$, the Rayleigh quotient of vector \f$v\f$ is

    \f[
       R(M, v) = \frac{v^\top M v}{v^\top v}.
    \f]

    @param mat a sparse matrix \f$M \in \mathbb{R}^{n \times n}\f$.
    @param vec a vector \f$v \in \mathbb{R}^n\f$.
    @return the Rayleigh quotient \f$R(M, v)\f$.
    """
    return stag_internal.rayleigh_quotient(utility.scipy_to_swig_sprs(mat),
                                           vec)



"""
Read and write graphs to disk.
"""
from . import stag_internal
from . import graph


@graph.return_graph
def load_edgelist(filename: str) -> graph.Graph:
    """
    Load a graph from an edgelist file.

    We define an edgelist file in the following way.
      - Any lines beginning with '#' or '//' are ignored
      - Any blank lines are ignored
      - All other lines are of one of the formats
          - `<u>, <v>, <weight>`
          - `<u>, <v>`
          - `<u> <v> <weight>`
          - `<u> <v>`

        where `<u>` and `<v>` can be parsed as integers, and `<weight>` can be parsed
        as a double. If the weight is omitted, it is taken to be 1.

    Here is an example edgelist file.

    \code
    # This line is ignored
    0, 1, 0.5
    1, 2, 1
    2, 0, 0.5
    \endcode

    @param filename the name of the edgelist file to be loaded
    @return stag.graph.Graph object
    @throws runtime_error if the file doesn't exist or cannot be parsed as an edgelist
    """
    return stag_internal.load_edgelist(filename)


def save_edgelist(g: graph.Graph, filename: str):
    """
    Save a graph as an edgelist file.

    @param g the graph object to be saved
    @param filename the name of the file to save the edgelist data to
    """
    stag_internal.save_edgelist(g.internal_graph, filename)


@graph.return_graph
def load_adjacencylist(filename: str) -> graph.Graph:
    r"""
     Load a graph from an adjacencylist file.

    The adjacency list file format is defined in the following way.
      - Any lines beginning with `#` or `//` are ignored
      - Any blank lines are ignored
      - All other lines have the format `<node_id>: <list of neighbours>`, where
        `<node_id>` is an integer and `<list of neighbours>` is either a
        space-separated list of integers or a space-separated list of
        `<id>:<weight>` where `<id>` gives the id of the neighbour and
        `<weight>` is the weight of the edge.
      - The `<node IDs>` of each line must be sorted in increasing order.
      - The graph should have no self-loops.

    Here is an example adjacencylist file.

        # This line is ignored
        0: 1 2
        1: 0 2 3
        2: 0 1
        3: 1

    The following example includes weighted edges.

        # This line is ignored
        0: 1:0.5 2:0.5
        1: 0:0.5 2:1
        2: 0:0.5 1:1

    Note that this method loads the entire graph into memory. For large graphs,
    the stag.graph.AdjacencyListLocalGraph object can be used to access the graph
    in a 'local' way without reading the entire graph into memory.

    @param filename the name of the adjacency list file to be loaded
    @return stag.graph.Graph object
    @throws runtime_error if the file doesn't exist or cannot be parsed as
            an adjacency list
    """
    return stag_internal.load_adjacencylist(filename)


def save_adjacencylist(g: graph.Graph, filename: str):
    r"""
    Save a graph as an adjacency list file.

    @param g the graph object to be saved.
    @param filename the name of the file to save the adjacency list data to.
    """
    stag_internal.save_adjacencylist(g.internal_graph, filename)


def edgelist_to_adjacencylist(edgelist_fname: str, adjacencylist_fname: str):
    r"""
    Convert an edgelist file to an adjacency list.

    @param edgelist_fname the name of the file containing the edgelist.
    @param adjacencylist_fname the name of the file to write the adjacency list.
    """
    stag_internal.edgelist_to_adjacencylist(edgelist_fname, adjacencylist_fname)


def adjacencylist_to_edgelist(adjacencylist_fname: str, edgelist_fname: str):
    r"""
    Convert an adjacency list file to an edgelist.

    @param adjacencylist_fname the name of the file containing the adjacency list.
    @param edgelist_fname the name of the file to write the edgelist.
    """
    stag_internal.adjacencylist_to_edgelist(adjacencylist_fname, edgelist_fname)



"""
Construct graphs from random models.
"""
import numpy as np
from typing import List
from . import stag_internal
from . import graph
from . import utility


@graph.return_graph
def sbm(n: int, k: int, p: float, q: float, exact: bool = False) -> graph.Graph:
    r"""
    Generate a graph from the symmetric stochastic block model.

    Every cluster has the same number of vertices. For large enough values of
    \f$n\f$, this method samples from an approximate stochastic block model by
    default which significantly speeds up the execution time.
    To sample exactly from the stochastic block model, pass the optional 'exact'
    parameter to the method.

    The approximate sampling method has running time \f$O(k^2 + \mathrm{nnz})\f$
    and the exact method has running time \f$O(n^2)\f$ where \f$\mathrm{nnz}\f$
    is the number of non-zeros in the generated graph's adjacency matrix.

    @param n the number of vertices in the graph.
    @param k the number of clusters. Vertices are split evenly between clusters
    @param p the probability of including an edge inside a cluster.
    @param q the probability of including an edge between two clusters.
    @param exact (optional) whether to use the exact probability distribution
                  or an approximation.
    @return the randomly generated stag.graph.Graph
    """
    return stag_internal.sbm(n, k, p, q, exact)


@graph.return_graph
def general_sbm(cluster_sizes: List[int],
                probabilities: np.ndarray,
                exact: bool = False) -> graph.Graph:
    r"""
    Generate a graph from the general stochastic block model.

    The `cluster_sizes` list specifies the number of vertices in each
    generated cluster.
    Let \f$k\f$ be the length of the cluster_sizes list.

    Then, `probabilities` should be a \f$k \times k\f$ matrix which specifies
    the edge probability between every pair of vertices.
    That is, for each pair of vertices \f$u\f$ and \f$v\f$, the probability of
    including the edge \f$\{u, v\}\f$ in the graph is \f$P(u, v)\f$, where
    \f$P\f$ is the `probabilities` matrix.

    The approximate sampling method has running time \f$O(k^2 + \mathrm{nnz})\f$
    where \f$\mathrm{nnz}\f$ is the number of non-zeros in the generated
    graph's adjacency matrix,
    and the exact
    method has running time \f$O(n^2)\f$.

    \par Example

    \code{python}
    import numpy as np
    import stag.graph
    import stag.random

    cluster_sizes = [100, 20, 10]
    prob_mat = np.asarray([[0.4, 0.1, 0.1],
                           [0.1, 0.7, 0],
                           [0.1, 0, 1]])
    my_graph = stag.random.general_sbm(cluster_sizes, prob_mat)
    print(my_graph.adjacency())
    \endcode

    @param cluster_sizes a list of length \f$k\f$ with the number of vertices
                         in each cluster.
    @param probabilities a \f$k \times k\f$ numpy matrix with the inter-cluster
                         probabilities.
    @param exact (optional) whether to use the exact probability distribution.
                 Default: false.
    @return the randomly generated graph
    """
    return stag_internal.general_sbm(stag_internal.vectorl(cluster_sizes),
                                     probabilities.astype(float),
                                     exact)


def general_sbm_edgelist(filename: str,
                         cluster_sizes: List[int],
                         probabilities: np.ndarray,
                         exact: bool = False):
    r"""
    Generate a graph from the general stochastic block model and save the
    resulting graph as an edgelist file.

    This method uses only constant memory since the graph can be streamed to
    disk while it is being generated.

    @param filename the edgelist file to save the graph to
    @param cluster_sizes a list of length \f$k\f$ with the number of vertices in
                         each cluster.
    @param probabilities a \f$k \times k\f$ matrix with the inter-cluster
                         probabilities.
    @param exact (optional) whether to use the exact probability distribution.
                 Default: false.
    """
    cluster_sizes = utility.possibly_convert_ndarray(cluster_sizes)
    return stag_internal.general_sbm_edgelist(filename,
                                              stag_internal.vectorl(cluster_sizes),
                                              probabilities.astype(float),
                                              exact)


@graph.return_graph
def erdos_renyi(n: int, p: float, exact: bool = False) -> graph.Graph:
    r"""
    Generate a graph from the Erdos-Renyi model.

    For large values of \f$n\f$, this method will use an approximate version of the
    random model with running time \f$O(\mathrm{nnz})\f$ where \f$\mathrm{nnz}\f$
    is the number of edges in the sampled graph.

    If the ``exact`` parameter is true, then the true Erdos-Renyi distribution
    will be used, with running time \f$O(n^2)\f$.

    @param n the number of vertices in the graph.
    @param p the probability of including each edge.
    @param exact (optional) whether to sample from the exact model.
    @return the randomly generated stag.graph.Graph
    """
    return stag_internal.erdos_renyi(n, p, exact)


def sbm_gt_labels(n: int, k: int) -> List[int]:
    r"""
    Construct a vector with the ground truth labels for a graph drawn from the
    symmetric stochastic block model.

    \par Example

    \code{python}
    import stag.graph
    import stag.random

    n = 6
    k = 3
    myGraph = stag.random.sbm(n, k, 0.8, 0.1)

    gt_labels = stag.random.sbm_gt_labels(n, k)

    # gt_labels is the list [0, 0, 1, 1, 2, 2].
    \endcode

    @param n the number of vertices in the graph
    @param k the number of clusters
    @return a list of integers containing the ground truth labels for the
            vertices in the graph.
    """
    return list(stag_internal.sbm_gt_labels(n, k))


def general_sbm_gt_labels(cluster_sizes: List[int]) -> List[int]:
    r"""
    Construct a vector with the ground truth labels for a graph drawn from the
    general stochastic block model.

    \par Example

    \code{python}
    import stag.graph
    import stag.random

    cluster_sizes = [4, 2]
    gt_labels = stag.random.general_sbm_gt_labels(cluster_sizes)

    # gt_labels is the list [0, 0, 0, 0, 1, 1]
    \endcode

    @param cluster_sizes a list of length \f$k\f$ with the number of vertices
                         in each cluster.
    @return a vector containing the ground truth labels for the vertices in the
            graph.
    """
    return list(stag_internal.general_sbm_gt_labels(stag_internal.vectorl(cluster_sizes)))



from . import stag_internal
import scipy.sparse
import inspect
import numpy as np

##
# \cond
# This file is currently listed as an explicit exception in the doxyfile.
# It will not be processed for documentation at all.
##

def swig_sprs_to_scipy(swig_mat):
    """
    Take a swig sparse matrix and convert it to a scipy sparse matrix.
    """
    outer_starts = stag_internal.sprsMatOuterStarts(swig_mat)
    inner_indices = stag_internal.sprsMatInnerIndices(swig_mat)
    values = stag_internal.sprsMatValues(swig_mat)
    return scipy.sparse.csc_matrix((values, inner_indices, outer_starts))


def scipy_to_swig_sprs(scipy_mat: scipy.sparse.csc_matrix):
    """
    Take a scipy sparse matrix and convert it to a swig sprs matrix.
    """
    col_starts = stag_internal.vectorl(scipy_mat.indptr.tolist())
    row_indices = stag_internal.vectorl(scipy_mat.indices.tolist())
    values = stag_internal.vectord(scipy_mat.data.tolist())
    return stag_internal.sprsMatFromVectors(col_starts,
                                            row_indices,
                                            values)


def return_sparse_matrix(func):
    def decorated_function(*args, **kwargs):
        swig_sparse_matrix = func(*args, **kwargs)
        sp_sparse = swig_sprs_to_scipy(swig_sparse_matrix)
        del swig_sparse_matrix
        return sp_sparse

    # Set the metadata of the returned function to match the original.
    # This is used when generating the documentation
    decorated_function.__doc__ = func.__doc__
    decorated_function.__module__ = func.__module__
    decorated_function.__signature__ = inspect.signature(func)

    return decorated_function


def possibly_convert_ndarray(argument):
    """
    Check whether argument is a numpy ndarray.
    If it is, convert it to a list. Otherwise return it as-is.
    """
    if isinstance(argument, np.ndarray):
        return argument.tolist()
    else:
        return argument


def convert_ndarrays(func):
    """A decorator for methods which take lists as arguments.
    Converts *all* ndarrays in the arguments to lists."""
    def decorated_function(*args, **kwargs):
        new_args = [possibly_convert_ndarray(arg) for arg in args]
        new_kwargs = {k: possibly_convert_ndarray(v) for k, v in kwargs.items()}
        return func(*new_args, **new_kwargs)


    # Set the metadata of the returned function to match the original.
    # This is used when generating the documentation
    decorated_function.__doc__ = func.__doc__
    decorated_function.__module__ = func.__module__
    decorated_function.__signature__ = inspect.signature(func)

    return decorated_function

##
# \endcond
##



"""
Graph object definitions and standard constructors.
"""
import networkx
from abc import ABC, abstractmethod
from typing import List
import inspect

import scipy.sparse

from . import stag_internal
from . import utility


class Edge(object):
    """
    An object representing a weighted edge in a graph.
    """

    def __init__(self, v1: int, v2: int, weight: float):
        """
        Create an edge from two node ids and a weight.
        """
        ## The ID of the first vertex in the edge.
        self.v1 = v1

        ## The ID of the second vertex in the edge.
        self.v2 = v2

        ## The weight of the edge
        self.weight = weight

        ##
        # \cond
        ##
        self.internal_edge = stag_internal.edge()
        self.internal_edge.v1 = v1
        self.internal_edge.v2 = v2
        self.internal_edge.weight = weight
        ##
        # \endcond
        ##

    ##
    # \cond
    ##

    def __eq__(self, other):
        return self.v1 == other.v1 and self.v2 == other.v2 and self.weight == other.weight

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return f"Edge({self.v1}, {self.v2}, weight={self.weight})"

    ##
    # \endcond
    ##


class LocalGraph(ABC):
    """
    \brief An abstract class which defines methods for exploring the
    local neighborhood of vertices in a graph.

    To maximise the performance of the local algorithms using this class,
    subclasses should cache the results of expensive queries. For example,
    if querying the neighbors of a vertex requires accessing the disk, then
    the result should be cached.
    """

    def __init__(self):
        r"""
        Default LocalGraph constructor.

        The constructor of any classes inheriting from this base class should
        call this constructor. For example:

        \code{python}
        import stag.graph
        class MyLocalGraph(stag.graph.LocalGraph):
          def __init__(self):
          super().__init__()

          # Add custom initialisation for MyLocalGraph
        \endcode
        """
        ##
        # \cond
        ##
        self.internal_graph = _PythonDefinedLocalGraph(self)
        ##
        # \endcond
        ##

    @abstractmethod
    def degree(self, v: int) -> float:
        """
        Given a vertex v, return its weighted degree.
        """
        pass

    @abstractmethod
    def degree_unweighted(self, v: int) -> int:
        """
        Given a vertex v, return its unweighted degree. That is, the number of
        neighbors of v, ignoring the edge weights.
        """
        pass

    @abstractmethod
    def neighbors(self, v: int) -> List[Edge]:
        """
        Given a vertex v, return a list of edges representing the
        neighbors of v.

        The returned edge objects will all have the ordering ``(v1, v2)`` such that
        ``edge.v1 = v``.

        @param v the ID of some vertex in the graph
        @return a list of stag.graph.Edge objects containing the neighbourhood
                of v
        """
        pass

    @abstractmethod
    def neighbors_unweighted(self, v: int) -> List[int]:
        """
        Given a vertex v, return a list of neighbors of v.

        The weights of edges to the neighbors are not returned by this method.

        @param v the ID of some vertex in the graph
        @return a list of vertex IDs giving the neighbours of v
        """
        pass

    def degrees(self, vertices: List[int]) -> List[float]:
        """
        Given a list of vertices, return a list of their weighted degrees.

        When developing implementations of the stag.graph.LocalGraph class,
        providing an efficient method of returning a list of degrees will
        improve the performance of local clustering algorithms.

        @param vertices a list of IDs representing the vertices to be queried
        @return a list of degrees
        """
        return [self.degree(v) for v in vertices]

    def degrees_unweighted(self, vertices: List[int]) -> List[int]:
        """
        Given a list of vertices, return a list of their unweighted degrees.

        When developing implementations of the stag.graph.LocalGraph class,
        providing an efficient method of returning a list of degrees will
        improve the performance of local clustering algorithms.

        @param vertices a list of IDs representing the vertices to be queried
        @return a list of unweighted degrees
        """
        return [self.degree_unweighted(v) for v in vertices]

    @abstractmethod
    def vertex_exists(self, v: int) -> bool:
        r"""
        Given a vertex ID, returns true or false to indicate whether the vertex exists
        in the graph.

        @param v the vertex index to check
        @return a boolean indicating whether there exists a vertex with the given index
        """
        pass

##
# \cond
# Do not document python defined local graph
##

class _PythonDefinedLocalGraph(stag_internal.LocalGraph):
    def __init__(self, python_local_graph: LocalGraph):
        super().__init__()
        self.python_graph = python_local_graph

    def degree(self, v: int):
        return self.python_graph.degree(v)

    def degree_unweighted(self, v: int):
        return self.python_graph.degree_unweighted(v)

    def neighbors(self, v: int):
        return [e.internal_edge for e in self.python_graph.neighbors(v)]

    def neighbors_unweighted(self, v: int):
        return self.python_graph.neighbors_unweighted(v)

    def degrees(self, vertices: List[int]):
        return self.python_graph.degrees(vertices)

    def degrees_unweighted(self, vertices):
        return self.python_graph.degrees_unweighted(vertices)

    def vertex_exists(self, v: int):
        return self.python_graph.vertex_exists(v)

##
# \endcond
##


class AdjacencyListLocalGraph(LocalGraph):
    """
    \brief A local graph backed by an adjacency list file on disk.

    The graph is loaded into memory in a local way only. That is, an adjacency
    list data structure is constructed in memory as node neighbours are queried.
    If a node is not found in the cached adjacency list, then the neighbours of
    the node are queried from the adjacency list on disk. This allows for local
    algorithms to be executed on very large graphs stored on disk without
    loading the whole graph into memory.

    See [Graph File Formats](@ref file-formats) for more information about the
    adjacency list file format.

    \note
    It is important that the adjacency list on disk is stored with sorted node
    indices. This allows us to query the neighbours of a given node in
    \f$O(\mathrm{log}(n))\f$ time using binary search.
    """

    def __init__(self, filename: str):
        r"""
        Construct a local graph backed by an adjacency list file.

        The adjacency list file must not be modified externally while it is in
        use by this object.

        @param filename the name of the adjacency list file which defines the
                        graph.
        """
        # Call the LocalGraph initialisation method - it is important that this
        # is called first. This is because we override the internal_graph
        # object in the current constructor.
        super().__init__()

        ##
        # \cond
        # Do not document the internal implementation of the object
        ##

        # This class is a thin wrapper around the stag_internal library, written in C++.
        # Initialise the internal graph object.
        self.internal_graph: stag_internal.AdjacencyListLocalGraph = \
            stag_internal.AdjacencyListLocalGraph(filename)

        ##
        # \endcond
        ##

    def degree(self, v: int) -> float:
        return self.internal_graph.degree(v)

    def degree_unweighted(self, v: int) -> int:
        return self.internal_graph.degree_unweighted(v)

    def neighbors(self, v: int) -> List[Edge]:
        return self.internal_graph.neighbors(v)

    def neighbors_unweighted(self, v: int) -> List[int]:
        return self.internal_graph.neighbors_unweighted(v)

    def vertex_exists(self, v: int) -> bool:
        return self.internal_graph.vertex_exists(v)



class Graph(LocalGraph):
    """
    \brief The core object used to represent graphs for use with the library.

    Graphs are always constructed from sparse matrices, and this is the internal
    representation used as well.
    Vertices of the graph are always referred to by their unique integer index.
    This index corresponds to the position of the vertex in the stored adjacency
    matrix of the graph.
    """

    def __init__(self, adj_mat: scipy.sparse.spmatrix,
                 internal_graph: stag_internal.Graph = None):
        r"""
        Initialise the graph with an adjacency matrix.
        
        For example:

        \code{python}
        >>> import stag.graph
        >>> import scipy.sparse
        >>>
        >>> adj_mat = scipy.sparse.csc_matrix([[0, 1, 1, 1],
        ...                                    [1, 0, 1, 1],
        ...                                    [1, 1, 0, 1],
        ...                                    [1, 1, 1, 0]])
        >>> g = stag.graph.Graph(adj_mat)
        \endcode

        @param adj_mat A sparse scipy matrix, such as ``scipy.sparse.csc_matrix``.
        @param internal_graph (optional) specify a STAG C++ graph object to
                              initialise with. Use this only if you understand
                              the internal workings of the STAG library.
        """
        # Call the LocalGraph initialisation method - it is important that this
        # is called first. This is because we override the internal_graph
        # object in the current constructor.
        super().__init__()

        ##
        # \cond
        # Do not document the internal implementation of the graph object
        ##

        # This class is essentially a thin wrapper around the stag_internal library, written in C++.
        if internal_graph is None:
            # Initialise the internal graph object with the provided adjacency matrix.
            adj_mat_csr = adj_mat.tocsr()
            outer_starts = stag_internal.vectorl(adj_mat_csr.indptr.tolist())
            inner_indices = stag_internal.vectorl(adj_mat_csr.indices.tolist())
            values = stag_internal.vectord(adj_mat_csr.data.tolist())
            self.internal_graph: stag_internal.Graph = stag_internal.Graph(outer_starts, inner_indices, values)
        else:
            # The initialiser was called with an internal graph object.
            self.internal_graph: stag_internal.Graph = internal_graph

        ##
        # \endcond
        ##

    @utility.return_sparse_matrix
    def adjacency(self) -> scipy.sparse.csc_matrix:
        """
        Return the sparse adjacency matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the graph adjacency matrix
        """
        return self.internal_graph.adjacency()

    @utility.return_sparse_matrix
    def laplacian(self) -> scipy.sparse.csc_matrix:
        """
        Construct the Laplacian matrix of the graph.

        The Laplacian matrix is defined by

        \f[
            L = D - A
        \f]

        where \f$D\f$ is the diagonal matrix of vertex degrees
        and \f$A\f$ is the adjacency matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the graph Laplacian
        """
        return self.internal_graph.laplacian()

    @utility.return_sparse_matrix
    def normalised_laplacian(self) -> scipy.sparse.csc_matrix:
        r"""
        Construct the normalised Laplacian matrix of the graph.

        The normalised Laplacian matrix is defined by

        \f[
            \mathcal{L} = D^{-1/2} L D^{-1/2}
        \f]

        where \f$D\f$ is the diagonal matrix of vertex degrees and \f$L\f$
        is the Laplacian matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the normalised Laplacian
        """
        return self.internal_graph.normalised_laplacian()

    @utility.return_sparse_matrix
    def normalised_laplacian(self) -> scipy.sparse.csc_matrix:
        r"""
        Construct the normalised Laplacian matrix of the graph.

        The normalised Laplacian matrix is defined by

        \f[
            \mathcal{L} = D^{-1/2} L D^{-1/2}
        \f]

        where \f$D\f$ is the diagonal matrix of vertex degrees and \f$L\f$
        is the Laplacian matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the normalised Laplacian
        """
        return self.internal_graph.normalised_laplacian()

    @utility.return_sparse_matrix
    def signless_laplacian(self) -> scipy.sparse.csc_matrix:
        """
        Construct the signless Laplacian matrix of the graph.

        The signless Laplacian matrix is defined by

        \f[
            J = D + A
        \f]

        where \f$D\f$ is the diagonal matrix of vertex degrees
        and \f$A\f$ is the adjacency matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the signless graph Laplacian
        """
        return self.internal_graph.signless_laplacian()

    @utility.return_sparse_matrix
    def normalised_signless_laplacian(self) -> scipy.sparse.csc_matrix:
        r"""
        Construct the normalised signless Laplacian matrix of the graph.

        The normalised signless Laplacian matrix is defined by

        \f[
            \mathcal{J} = D^{-1/2} J D^{-1/2}
        \f]

        where \f$D\f$ is the diagonal matrix of vertex degrees and \f$J\f$
        is the signless Laplacian matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the normalised signless Laplacian
        """
        return self.internal_graph.normalised_signless_laplacian()

    @utility.return_sparse_matrix
    def degree_matrix(self) -> scipy.sparse.csc_matrix:
        r"""
        The degree matrix of the graph.

        The degree matrix \f$D \in \mathbb{R}^{n \\times n}\f$
        is a diagonal matrix such that \f$D(i, i) = \mathrm{deg}(i)\f$
        where \f$\mathrm{deg}(i)\f$ is the degree of vertex \f$i\f$ and
        \f$n\f$ is the number of vertices in the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the degree matrix
        """
        return self.internal_graph.degree_matrix()
    
    @utility.return_sparse_matrix
    def inverse_degree_matrix(self) -> scipy.sparse.csc_matrix:
        r"""
        The inverse degree matrix of the graph.

        The degree matrix \f$D^{-1} \in \mathbb{R}^{n \times n}\f$
        is a diagonal matrix such that

        \f[
            D(i, i) =  \left\{
                    \begin{array}{lll}
                        \mathrm{deg}(i) & \mbox{if } \mathrm{deg}(i) > 0 \\
                        0 & \mbox{otherwise}
                    \end{array}
                \right.
        \f]

        where \f$\mathrm{deg}(i)\f$ is the degree of vertex \f$i\f$ and
        \f$n\f$ is the number of vertices in the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the inverse degree matrix
        """
        return self.internal_graph.inverse_degree_matrix()
    
    @utility.return_sparse_matrix
    def lazy_random_walk_matrix(self) -> scipy.sparse.csc_matrix:
        """
        The lazy random walk matrix of the graph.

        The lazy random walk matrix is defined by

        \f[
            W = \frac 1 2 I + \frac 1 2 A D^{-1}
        \f]

        where \f$I\f$ is the identity matrix, \f$A\f$ is the graph adjacency
        matrix and \f$D\f$ is the degree matrix of the graph.

        @return a ``scipy.sparse.csc_matrix`` representing the lazy random walk
                matrix
        """
        return self.internal_graph.lazy_random_walk_matrix()

    def total_volume(self) -> float:
        r"""
        The volume of the graph.

        The volume of a graph \f$G = (V, E, w)\f$ is defined by

        \f[
            \mathrm{vol}(G) = \sum_{u \in V} \mathrm{deg}(u),
        \f]

        where \f$\mathrm{deg}(u)\f$ is the degree of vertex \f$u\f$.

        @return the graph's volume, \f$\mathrm{vol}(G)\f$
        """
        return self.internal_graph.total_volume()

    def average_degree(self) -> float:
        r"""
        The average degree of the graph.

        This is defined as the sum of the node degrees divided by the number of nodes.

        @return the graph's average degree.
        """
        return self.internal_graph.average_degree()

    def number_of_vertices(self) -> int:
        """The number of vertices in the graph."""
        return self.internal_graph.number_of_vertices()

    def number_of_edges(self) -> int:
        """
        The number of edges in the graph.

        This method ignores the weights of the edges.
        """
        return self.internal_graph.number_of_edges()

    def degree(self, v: int) -> float:
        return self.internal_graph.degree(v)

    def degree_unweighted(self, v: int) -> int:
        return self.internal_graph.degree_unweighted(v)

    def neighbors(self, v: int) -> List[Edge]:
        return self.internal_graph.neighbors(v)

    def neighbors_unweighted(self, v: int) -> List[int]:
        return self.internal_graph.neighbors_unweighted(v)

    def vertex_exists(self, v: int) -> bool:
        return self.internal_graph.vertex_exists(v)

    ##
    # \cond
    # Do not document the eq method for the graph object.
    ##
    def __eq__(self, other):
        #
        # IMPORTANT NOTE
        # Checking whether two graphs are the same is thought to be a 'difficult'
        # problem, and so this method does not really check for mathematical equivalence.
        # Rather, we just check that the adjacency matrices are the same. As such, this method
        # should not be relied on to test for graph isomorphism!
        #

        # Check basic size information about the two graphs first
        if self.number_of_vertices() != other.number_of_vertices():
            return False

        if self.number_of_edges() != other.number_of_edges():
            return False

        # Check that the data vectors of the graph adjacency matrices are equal.
        a1 = self.internal_graph.adjacency()
        a2 = other.internal_graph.adjacency()
        if stag_internal.sprsMatOuterStarts(a1) != stag_internal.sprsMatOuterStarts(a2):
            return False
        if stag_internal.sprsMatInnerIndices(a1) != stag_internal.sprsMatInnerIndices(a2):
            return False
        return stag_internal.sprsMatValues(a1) == stag_internal.sprsMatValues(a2)

    def __ne__(self, other):
        return not self.__eq__(other)

    ##
    # \endcond
    ##

    def to_networkx(self) -> networkx.Graph:
        """
        Construct a networkx graph object which is equivalent to this STAG graph.

        See the
        [networkx documentation](https://networkx.org/documentation/stable/reference/classes/graph.html).
        """
        return networkx.Graph(self.adjacency())

    def draw(self, **kwargs):
        """
        Plot the graph with matplotlib.

        This uses the networkx draw method and accepts any the keyword arguments
        will be passed through directly.

        \par Example

        \code{python}
        import matplotlib.pyplot as plt
        import stag.graph
        myGraph = stag.graph.star_graph(10)
        myGraph.draw()
        plt.show()
        \endcode
        """
        netx_graph = self.to_networkx()
        networkx.draw(netx_graph, **kwargs)

##
# \cond
# A decorator which transforms a graph returned from the C++ library to the
# python Graph object.
##
def return_graph(func):
    def decorated_function(*args, **kwargs):
        swig_graph = func(*args, **kwargs)
        return Graph(None, internal_graph=swig_graph)

    # Set the metadata of the returned function to match the original.
    # This is used when generating the documentation
    decorated_function.__doc__ = func.__doc__
    decorated_function.__module__ = func.__module__
    decorated_function.__signature__ = inspect.signature(func)

    return decorated_function

##
# \endcond
##

@return_graph
def cycle_graph(n) -> Graph:
    """
    Construct a cycle graph on n vertices.

    @param n the number of vertices in the constructed graph
    @return a stag.graph.Graph object representing a cycle graph
    """
    return stag_internal.cycle_graph(n)


@return_graph
def complete_graph(n) -> Graph:
    """
    Construct a complete graph on n vertices.

    @param n the number of vertices in the constructed graph
    @return a stag.graph.Graph object representing a complete graph
    """
    return stag_internal.complete_graph(n)


@return_graph
def barbell_graph(n) -> Graph:
    """
    Construct a barbell graph. The barbell graph consists of 2 cliques on n
    vertices, connected by a single edge.

    @param n the number of vertices in each of the two cliques.
             The returned graph will have \f$2n\f$ vertices.
    @return a stag.graph.Graph object representing the barbell graph
    """
    return stag_internal.barbell_graph(n)


@return_graph
def star_graph(n) -> Graph:
    """
    Construct a star graph. The star graph consists of one central vertex
    connected by an edge to n-1 outer vertices.

    @param n the number of vertices in the constructed graph
    @return a stag.graph.Graph object representing the star graph
    """
    return stag_internal.star_graph(n)


def from_networkx(netx_graph: networkx.Graph,
                  edge_weight_attribute: str = "weight"):
    """
    Given a networkx graph, convert it to a stag.graph.Graph object.

    Unless otherwise specified, this method will use the 'weight' attribute on the
    networkx edges to assign the weight of the edges. If no such attribute is present,
    the edges will be added with weight \f$1\f$.

    @param netx_graph The networkx graph object to be converted.
    @param edge_weight_attribute (default 'weight') the edge attribute to be used to
                                  generate the weights
    @return A stag.graph.Graph object which is equivalent to the networkx graph.
    """
    return Graph(networkx.adjacency_matrix(netx_graph,
                                           weight=edge_weight_attribute))



"""Tests for the clustering algorithms."""
import scipy.sparse
import pytest
import math
import numpy as np
from context import stag
import stag.graph
import stag.cluster
import stag.random
import stag.spectrum

def test_power_method():
    g = stag.graph.complete_graph(3)
    lap = g.laplacian()
    assert(lap.shape == (3, 3))
    vec = np.asarray([[0, 1, 0]]).T
    assert vec.shape == (3, 1)

    newvec = stag.spectrum.power_method(lap,
                                        num_iterations=2,
                                        initial_vector=vec)

    assert np.allclose(newvec, np.asarray([[-1/math.sqrt(6),
                                             2/math.sqrt(6),
                                            -1/math.sqrt(6)]]).T, atol=0.00001)

    g = stag.graph.cycle_graph(10)
    lap = g.normalised_laplacian()
    dominant_eigvec = stag.spectrum.power_method(lap, num_iterations=1000)
    assert abs(2 - stag.spectrum.rayleigh_quotient(lap, dominant_eigvec)) < 0.0001


def test_eigensystem():
    g = stag.graph.complete_graph(4)
    lap = g.laplacian()
    eigval, eigvec = stag.spectrum.compute_eigensystem(lap, 1)
    assert np.allclose(eigval, np.asarray([0]))
    assert np.allclose(eigvec, np.asarray([[1/2, 1/2, 1/2, 1/2]]).T)



"""
Interface to a neo4j database, exposing a graph interface on which we can run
STAG algorithms.
"""
import neo4j
from typing import List
from functools import lru_cache

from . import graph


class Neo4jGraph(graph.LocalGraph):
    """
    Represent a neo4j database as a stag.graph.LocalGraph object.

    This provides 'local' access to the graph only. Given a node, we can query
    its neighbors and its degree.
    The graph does not support weighted edges, and it does not
    distinguish between edge types or directions.

    The neo4j ``<id>`` attributes are used as the STAG vertex ids.

    All methods which make calls to the database backend have a small cache of
    recent queries, meaning that consecutive queries for the same node will be
    fast.
    """
    def __init__(self, uri: str, username: str, password: str):
        """
        Connect to a neo4j database backend.

        For example, assuming that there is a Neo4j database running locally,
        the graph object can be created as follows.

        \code{python}
        import stag.neo4j
        g = stag.neo4j.Neo4jGraph("bolt://localhost:7687", "neo4j", "password")
        \endcode

        It is also possible to connect to a remote Neo4j database by passing
        the relevant ``uri`` to the constructor.

        @param uri the location of the neo4j database
        @param username the neo4j username
        @param password the neo4j password
        """
        super().__init__()

        ## The neo4j database driver object used to query the underlying
        # database.
        self.driver = neo4j.GraphDatabase.driver(uri, auth=(username, password))

        ##
        # \cond
        # The partially-constructed adjacency list of all nodes we've queried
        # so far.
        ##
        self.adjacency_list = {}
        self.labels_cache = {}
        ##
        # \endcond
        ##

    def degree(self, v):
        """
        Equivalent to stag.neo4j.Neo4jGraph.degree_unweighted.
        """
        return self.degree_unweighted(v)

    def degree_unweighted(self, v: int) -> int:
        """Query the degree of the node with the given neo4j ID."""
        # The degree of a node is the length of its list of neighbours
        ns = self.neighbors_unweighted(v)
        return len(ns)

    def neighbors(self, v: int) -> List[graph.Edge]:
        """
        Fetch the neighbors of the node with the given Neo4j node ID.

        The returned stag.graph.Edge objects all have weight 1.
        """
        return [graph.Edge(v, u, 1) for u in self.neighbors_unweighted(v)]

    def neighbors_unweighted(self, v: int) -> List[int]:
        """
        Fetch the neighbors of the node with the given Neo4j node ID.

        Returns the Neo4j node IDs of the neighboring nodes.
        """
        if v not in self.adjacency_list:
            with self.driver.session() as session:
                result = session.execute_read(self._neighbors_query, v)
            self.adjacency_list[v] =  [x[0] for x in result]
        return self.adjacency_list[v]

    ##
    # \cond
    ##
    @staticmethod
    def _neighbors_query(tx, node_id):
        # To get the neighbors of the given node, we will execute the following
        # Cypher command:
        #    MATCH (n1)-[]-(n2) WHERE id(n1) = v return id(n2)
        # which finds the nodes which are a single step from the node
        # with the given node ID.
        result = tx.run("MATCH (n1)-[]-(n2) "
                        "WHERE id(n1) = $node_id "
                        "RETURN DISTINCT id(n2)", node_id=node_id)
        return list(result.values())
    ##
    # \endcond
    ##

    def query_node_labels(self, node_id: int) -> List[str]:
        """
        Query the labels of the given node.

        For example, using the Neo4j movie database example, you can query
        whether a given node represents a person or a movie as follows.

        \code{python}
            >>> import stag.neo4j
            >>> g = stag.neo4j.Neo4jGraph("bolt://localhost:7687", "neo4j", "password")
            >>> labels = g.query_node_labels(0)
            ['Movie']
            >>> labels = g.query_node_labels(1)
            ['Person']
        \endcode

        """
        if node_id not in self.labels_cache:
            with self.driver.session() as session:
                result = session.execute_read(self._labels_query, node_id)
            self.labels_cache[node_id] = [x[0] for x in result]
        return self.labels_cache[node_id]

    ##
    # \cond
    ##
    @staticmethod
    def _labels_query(tx, node_id: int):
        # To get the neighbors of the given node, we will execute the following
        # Cypher command:
        #    MATCH (n1)-[]-(n2) WHERE id(n1) = v return id(n2)
        # which finds the nodes which are a single step from the node
        # with the given node ID.
        result = tx.run("MATCH (n1) "
                        "WHERE id(n1) = $node_id "
                        "RETURN DISTINCT labels(n1)", node_id=node_id)
        return result.values()[0]
    ##
    # \endcond
    ##

    def query_id(self, property_name: str, property_value: str) -> int:
        """
        Find the Neo4j ID of a node with the given property.

        For example, using the Neo4j movie database example you can find the
        Neo4j node corresponding to a given movie as follows.
        
        \code{python}
            import stag.neo4j
            g = stag.neo4j.Neo4jGraph("bolt://localhost:7687", "neo4j", "password")
            node_id = g.query_id("title", "Avatar")
        \endcode

        This will make a query to the database to find a node with the given
        property.
        This query will be slow unless the database has an index for
        ``property_name``.
        For more information, see the Neo4j documentation on
        [creating an index](https://neo4j.com/docs/cypher-manual/current/indexes-for-search-performance/).

        @param property_name the node property to query
        @param property_value the value of the node property to search for
        @return the Neo4j node ID of a node with the given property value.
                Returns None if there is no node in the dataabase with the
                requested property.
        """
        with self.driver.session() as session:
            result = session.execute_read(self._id_query, property_name, property_value)
        return result

    ##
    # \cond
    ##
    @staticmethod
    def _id_query(tx, property_name, property_value):
        result = tx.run(f'MATCH (n1 {{{property_name}: "{property_value}"}}) '
                        "RETURN id(n1) "
                        "LIMIT 1")
        result = result.single()
        if result is None:
            return None
        else:
            return result[0]
    ##
    # \endcond
    ##

    def query_property(self, id: int, property_name: str) -> str:
        """
        Find the requested property of a Neo4j node.

        For example, using the Neo4j movie database example, you can find the
        title of a movie node as follows.

        \code{python}
            import stag.neo4j
            g = stag.neo4j.Neo4jGraph("bolt://localhost:7687", "neo4j", "password")
            title = g.query_property(1, "title")
        \endcode

        @param id the Neo4j node ID to query
        @param property_name the name of the property to query
        @return the value of the requested property on the given node. Returns
                None if the requested property is not set on the requested node,
                or if the node id does not exist.
        """
        with self.driver.session() as session:
            result = session.execute_read(self._property_query, id, property_name)
        return result

    @lru_cache(maxsize=1024)
    def vertex_exists(self, v: int) -> bool:
        with self.driver.session() as session:
            result = session.execute_read(self._exists_query, v)
        return result

    ##
    # \cond
    ##
    @staticmethod
    def _exists_query(tx, id: int):
        result = tx.run('MATCH (n1) '
                        f"WHERE id(n1) = {id} "
                        "RETURN n1 "
                        "LIMIT 1")
        result = result.single()
        if result is None:
            return False
        else:
            return True
    ##
    # \endcond
    ##

    ##
    # \cond
    ##
    @staticmethod
    def _property_query(tx, node_id, property_name):
        result = tx.run("MATCH (n1) "
                        f"WHERE id(n1) = {node_id} "
                        f"RETURN (n1.{property_name}) "
                        "LIMIT 1")
        result = result.single()
        if result is None:
            return None
        else:
            return result[0]

    def __del__(self):
        # When the graph object it destroyed, close the connection to the backing
        # database.
        self.driver.close()
    ##
    # \endcond
    ##



"""Algorithms for finding clusters in graphs."""
import scipy.sparse
from typing import List, Tuple

from . import stag_internal
from . import graph
from . import utility


def spectral_cluster(g: graph.Graph, k: int) -> List[int]:
    r"""
    Spectral clustering algorithm.

    This is a simple graph clustering method, which provides a clustering of the entire graph.
    To use spectral clustering, simply pass a `stag.graph.Graph` object
    and the number of clusters you would like to find.

    \code{python}
    import stag.graph
    import stag.cluster

    myGraph = stag.graph.Graph.barbell_graph(10)
    labels = stag.cluster.spectral_cluster(myGraph, 2)
    print(labels)
    \endcode

    The spectral clustering algorithm has the following steps.
      - Compute the \f$k\f$ smallest eigenvectors of the normalised Laplacian matrix.
      - Embed the vertices into \f$\mathbb{R}^k\f$ according to the eigenvectors.
      - Cluster the vertices into \f$k\f$ clusters using a \f$k\f$-means clustering algorithm.

    @param g the graph object to be clustered
    @param k the number of clusters to find. Should be less than \f$n/2\f$.
    @return a list of ints giving the cluster membership for each vertex in the graph

    \par References
    A. Ng, M. Jordan, Y. Weiss.
    On spectral clustering: Analysis and an algorithm. NeurIPS'01
    """
    return list(stag_internal.spectral_cluster(g.internal_graph, k))

def local_cluster(g: graph.LocalGraph, seed_vertex: int, target_volume: float) -> List[int]:
    r"""
    Local clustering algorithm based on personalised Pagerank.

    Given a graph and starting vertex, return a cluster which is close to the
    starting vertex.

    This method uses the ACL local clustering algorithm.

    @param g a graph object implementing the LocalGraph interface
    @param seed_vertex the starting vertex in the graph
    @param target_volume the approximate volume of the cluster you would like to find
    @return a vector containing the indices of vectors considered to be in the
            same cluster as the seed_vertex.

    \par References
    R. Andersen, F. Chung, K. Lang.
    Local graph partitioning using pagerank vectors. FOCS'06
    """
    return list(stag_internal.local_cluster(g.internal_graph, seed_vertex, target_volume))


def local_cluster_acl(g: graph.LocalGraph,
                      seed_vertex: int,
                      locality: float,
                      error: float = 0.001) -> List[int]:
    r"""
    The ACL local clustering algorithm. Given a graph and starting vertex,
    returns a cluster close to the starting vertex, constructed in a local way.

    The locality parameter is passed as the alpha parameter in the personalised
    pagerank calculation.

    @param g a graph object implementing the LocalGraph interface
    @param seed_vertex the starting vertex in the graph
    @param locality a value in \f$[0, 1]\f$ indicating how 'local' the cluster should
                    be. A value of \f$1\f$ will return the return only the seed vertex
                    and a value of \f$0\f$ will explore the whole graph.
    @param error (optional) - the acceptable error in the calculation of the approximate
                              pagerank. Default \f$0.001\f$.
    @return a vector containing the indices of vectors considered to be in the
            same cluster as the seed_vertex.

    \par References
    R. Andersen, F. Chung, K. Lang.
    Local graph partitioning using pagerank vectors. FOCS'06
    """
    return list(stag_internal.local_cluster_acl(g.internal_graph,
                                                seed_vertex,
                                                locality,
                                                error))


def approximate_pagerank(g: graph.LocalGraph,
                         seed_vector: scipy.sparse.csc_matrix,
                         alpha: float,
                         epsilon: float) -> Tuple[scipy.sparse.csc_matrix,
                                                  scipy.sparse.csc_matrix]:
    r"""
    Compute the approximate pagerank vector.

    The parameters s, alpha, and epsilon are used as described in the ACL paper.

    Note that the dimension of the returned vectors may not match the true
    number of vertices in the graph provided since the approximate
    pagerank is computed locally.

    @param g a stag.graph.LocalGraph object
    @param seed_vector the seed vector of the personalised pagerank
    @param alpha the locality parameter of the personalised pagerank
    @param epsilon the error parameter of the personalised pagerank
    @return A tuple of sparse column vectors corresponding to
             - p: the approximate pagerank vector
             - r: the residual vector

            By the definition of approximate pagerank, it is the case that
               p + ppr(r, alpha) = ppr(s, alpha).

    @throws argument_error if the provided seed_vector is not a column vector.

    \par References
    R. Andersen, F. Chung, K. Lang.
    Local graph partitioning using pagerank vectors. FOCS'06
    """
    apr = stag_internal.approximate_pagerank(g.internal_graph,
                                              utility.scipy_to_swig_sprs(seed_vector),
                                              alpha,
                                              epsilon)
    return utility.swig_sprs_to_scipy(apr[0]), utility.swig_sprs_to_scipy(apr[1])


def sweep_set_conductance(g: graph.LocalGraph, v: scipy.sparse.csc_matrix) -> List[int]:
    r"""
    Find the sweep set of the given vector with the minimum conductance.

    First, sort the vector such that \f$v_1, \ldots, v_n\f$. Then let

    \f[
        S_i = \{v_j : j <= i\}
    \f]

    and return the set of original indices corresponding to

    \f[
        \mathrm{argmin}_i \phi(S_i)
    \f]

    where \f$\phi(S)\f$ is the conductance of \f$S\f$.

    This method is expected to be run on vectors whose support is much less
    than the total size of the graph. If the total volume of the support of vec
    is larger than half of the volume of the total graph, then this method may
    return unexpected results.

    Note that the caller is responsible for any required normalisation of the
    input vector. In particular, this method does not normalise the vector by
    the node degrees.

    @param g a stag.graph.LocalGraph object
    @param v the vector to sweep over
    @return a vector containing the indices of vec which give the minimum
            conductance in the given graph
    """
    return stag_internal.sweep_set_conductance(g.internal_graph, utility.scipy_to_swig_sprs(v))


@utility.convert_ndarrays
def adjusted_rand_index(gt_labels: List[int], labels: List[int]) -> float:
    r"""
    Compute the Adjusted Rand Index between two label vectors.

    @param gt_labels the ground truth labels for the dataset
    @param labels the candidate labels whose ARI should be calculated
    @return the ARI between the two labels vectors

    \par References
    W. M. Rand.
    Objective criteria for the evaluation of clustering methods.
    Journal of the American Statistical Association. 66 (336): 846–850. 1971.
    """
    return stag_internal.adjusted_rand_index(stag_internal.vectorl(gt_labels),
                                             stag_internal.vectorl(labels))


@utility.convert_ndarrays
def conductance(g: graph.LocalGraph, cluster: List[int]) -> float:
    r"""
     Compute the conductance of the given cluster in a graph.

    Given a graph \f$G = (V, E)\f$, the conductance of \f$S \subseteq V\f$
    is defined to be

    \f[
       \phi(S) = \frac{w(S, V \setminus S)}{\mathrm{vol}(S)},
    \f]

    where \f$\mathrm{vol}(S) = \sum_{v \in S} \mathrm{deg}(v)\f$ is the volume
    of \f$S\f$ and \f$w(S, V \setminus S)\f$ is the total weight of edges crossing
    the cut between \f$S\f$ and \f$V \setminus S\f$.

    @param g a stag.graph.LocalGraph object representing \f$G\f$.
    @param cluster a list of node IDs in \f$S\f$.
    @return the conductance \f$\phi_G(S)\f$.
    """
    return stag_internal.conductance(g.internal_graph,
                                     stag_internal.vectorl(cluster))
